########################################################################
# Alsa Mixer Package
=encoding utf8

=head1 NAME

Audio::Alsa::Mixer -- control alsa audio devices

=head1 SYNOPSIS

After some problems contoling the Volume of Alsa Devices
(amixer is not easy to handle) i created this Package
In this state it's more or less an hack around "amixer"

The Hirachy is:

B<Audio::Alsa::Mixer> contains the Devices

=over 2

B<Audio::Alsa::Mixer::Device> contains the Controls

=over 2

B<Audio::Alsa::Mixer::Device::Control> contains the Channels

=back

=back

=head2 Example1

=over 4

 #!/usr/bin/perl
 use strict;
 use warnings;
 use Audio::Alsa::Mixer;

 for(Audio::Alsa::Mixer->device_list())
 {
   print $_->name()."\n";
   print "  ID         : ".($_->id()//'')."\n"; # id() is special because the default Mixer has no ID.
   print "  Name       : ".$_->name()."\n";
   print "  ID-Name    : ".$_->id_name()."\n";
   print "  Mixer-Name : ".$_->mixer_name()."\n";
   print "  Full-Name  : ".$_->full_name()."\n";
   print "  Components : ".$_->components()."\n";
   print "  Description: ".$_->description()."\n";

   for($_->control_list())
   {
     print "   ".$_->name()."\n";
     for($_->channel_list())
     {
       print "     ".$_->name()."\n";
       print "       GET OUT:".$_->volume_out('%')."%\n";
       print "       GET IN :".$_->volume_in('%')."%\n";
       print "       MUTE   :".($_->mute()?"yes":"no")."\n";
     }
   }
   print "#####################################################\n";
 }


=back

=head2 Example2

=over 4

 #!/usr/bin/perl
 use strict;
 use warnings;
 use Audio::Alsa::Mixer;

 print "Volume: ",
       Alsa::Mixer->device('USB-Audio')->control('Speaker')->channel('Front Left')->volume_out('%'),
       "%\n";

 print "Volume: ",
       Alsa::Mixer->device()->control('Master')->channel('Front Left')->volume_out('%'),
       "%\n";

=back

=head1 DESCRIPTION

Controll Alsa Audiodevices

=head1 AUTHOR

ToPeG

=head1 USING

alsa tool "amixer"

=head1 METHODS

=over 4

=cut

# CHANGELOG:
#
# 0.1 initial Release

use strict;
use warnings;

my $cards='/proc/asound/cards';

my $cmd='LANG=C /usr/bin/amixer';
my $devices_cmd  = "$cmd %s info 2>/dev/null";
my $controls_cmd = "$cmd %s scontrols  2>/dev/null";
my $channels_cmd = "$cmd %s scontents  2>/dev/null";
my $set_cmd      = "$cmd %s -q cset iface=MIXER,name='%s Volume' %s  >/dev/null 2>&1";
my $get_cmd      = "$cmd %s cget iface=MIXER,name='%s Volume'  2>/dev/null";
my $get_mute_cmd = "$cmd %s get '%s'";
my $set_mute_cmd = "$cmd %s -q set '%s' %s >/dev/null 2>&1";

########################################################################
########################################################################
# helper class usabele with
#  use base 'Audio::Alsa::Mixer::Helper';
package Audio::Alsa::Mixer::Helper;

sub _run_cmd
{
  my $self=shift;
  my $cmd=shift;
  my $ret=shift;
  if(open(my $ch, '-|' ,$cmd))
  {
    local $/=undef;
    $ret=<$ch>;
    close($ch);
  }
  else
  { warn("Error exec $cmd ($!)\n"); }
  return $ret;
}

sub _exec_cmd
{
  my $self=shift;
  my $cmd=shift;
  return system($cmd)==0;
}

sub _c_pre
{
  my $self=shift;
  return '' unless defined($_[0]);
  my $id=$self->{shift()};
  return defined($id)?"-c $id":'';
}

#get/set values
sub _value
{
  my $self=shift;
  my $name=shift;
  return undef unless(exists($self->{$name}));
  $self->{$name}=shift() if(@_);
  return $self->{$name};
}
########################################################################
########################################################################
# helper class usabele with
#  use base 'Audio::Alsa::Mixer::Helper';
package Audio::Alsa::Mixer::Helper::Audio;
use base 'Audio::Alsa::Mixer::Helper';

# get all volumes of the Control
# this is needed because one all Channels can be set/get at once.
sub _get_volume_list
{
  my $self=shift;
  my $type=shift;
  my $val=shift;
  $val='' unless($val && $val eq '%');

  my @list;

  my $name=$type;
  if($self->{control_name} ne $type)
  { $name="$self->{control_name} $type"; }

  my $cmd=sprintf($get_cmd, $self->_c_pre('device_id'), $name);
  my $data=$self->_run_cmd($cmd,'');

  if($data=~/:\s+values=([^\n]+)/)
  {
    @list=split(/,/,$1);
    if($val eq '%')
    { $_=int((100/$self->{$type}->{max})*$_) for(@list); }
  }

  return @list;
}

sub _set_volume_list
{
  my $self=shift;
  my $type=shift;
  return 0 unless($self->{$type} && ref($self->{$type}) eq 'HASH');
  my @vals=@_;

  for(@vals)
  {
    $_=$self->{$type}->{min} if($_<$self->{$type}->{min});
    $_=$self->{$type}->{max} if($_>$self->{$type}->{max});
  }

  my $name=$type;
  if($self->{control_name} ne $type)
  { $name="$self->{control_name} $type"; }

  my $cmd=sprintf($set_cmd, $self->_c_pre('device_id'), $name, join(',',@vals));

  return $self->_exec_cmd($cmd);
}
########################################################################
########################################################################

=back

=head2 Audio::Alsa::Mixer

=over 4

=cut

package Audio::Alsa::Mixer;

=item  Audio::Alsa::Mixer->device_list()

Get a list of all Devices controlled by Alsa

Special is the first Element.
it's the primary controlled Device, and can be an virtual Device
(when, as Example, PulseAudio is running)
otherwise it's a copy on a device in the list.

B<Return:>

List of I<Audio::Alsa::Mixer::Device> Objects

B<Example:>

 my @list=Audio::Alsa::Mixer->device_list();

=cut
sub device_list
{
  my @list;
  if(open(my $fh, '<', $cards))
  {
    local $/=undef;
    my $data=<$fh>;
    close($fh);

    my $default=Audio::Alsa::Mixer::Device->_new();
    push(@list,$default) if($default && $default->name());

    while($data=~/(\d+)\s+\[(.+?)\s*\]:\s*(\S+)\s+-\s+(.+?)\s*\n\s*(.+?)\n/gcs)
    { push(@list,Audio::Alsa::Mixer::Device->_new($1,$2,$3,$4,$5)); }
  }
  else
  { warn("ERROR open $cards ($!)\n"); }
  return @list;
}

=item  Audio::Alsa::Mixer->device(<id>)

Get one Device by Name or Id
If no Value is is set the Alsa-Default will returned

B<Return:>

one Audio::Alsa::Mixer::Device Object or undef

B<Example:>

 # get device with ID 0
 my $device=Audio::Alsa::Mixer->device(0);

=cut
sub device
{
  my $class=shift;
  my $id=shift;
  my @l=$class->device_list();

  return shift(@l) unless(defined($id));

  for(@l)
  {
    return $_ if($_->id()//''     eq $id);
    return $_ if($_->id_name()    eq $id);
    return $_ if($_->mixer_name() eq $id);
    return $_ if($_->name()       eq $id);
  }

  return undef;
}

########################################################################
########################################################################
=back

=head2 Audio::Alsa::Mixer::Device

=over 4

=cut
package Audio::Alsa::Mixer::Device;
use base 'Audio::Alsa::Mixer::Helper';

# create new object
sub _new
{
  my $class=shift;
  my $self=bless({},$class);
  $self->{id}         =shift();
  $self->{id_name}    =shift();
  $self->{name}       =shift();
  $self->{full_name}  =shift();
  $self->{description}=shift();

  $self->_read_info();

  return $self;
}

# read alsa device informations
sub _read_info
{
  my $self=shift;

  my $cmd=sprintf($devices_cmd, $self->_c_pre('id'));
  my $data=$self->_run_cmd($cmd,'');

  if($data=~m#Card hw:(\d+) '([^']+)'/'((.+?) at [^\n]+)'#)
  {
    $self->{id}         =$1 unless(defined($self->{id}));
    $self->{id_name}    =$2 unless(defined($self->{id_name}));
    $self->{name}       =$2 unless(defined($self->{name}));
    $self->{description}=$3 unless(defined($self->{description}));
    $self->{full_name}  =$4 unless(defined($self->{full_name}));
  }
  elsif($data=~m#Card default '([^']+)'/'(.+?)'#)
  {
    $self->{id}         =undef;
    $self->{id_name}    =$1 unless(defined($self->{id_name}));
    $self->{description}='System Default' unless(defined($self->{description}));
    $self->{full_name}  =$2 unless(defined($self->{full_name}));
    $self->{name}       ='default' unless(defined($self->{name}));
  }

  my %opts;
  while($data=~/(?:^|\n)\s*([^\n]+?)\s*:\s*(?:'([^']*)'|(\d+))\s*(?=\n|$)/gcs)
  {
    $opts{$1}=$2//$3;
    #print "'$1' = '$opts{$1}'\n";
  }

  $self->{mixer_name}=$opts{'Mixer name'}//'';
  $self->{components}=$opts{'Components'}//'';
}

=item  $device->id()

=cut
sub id          { shift->_value('id',         @_); }

=item  $device->id_name()

=cut
sub id_name     { shift->_value('id_name',    @_); }

=item  $device->components()

=cut
sub components  { shift->_value('components', @_); }

=item  $device->mixer_name()

=cut
sub mixer_name  { shift->_value('mixer_name', @_); }

=item  $device->name()

=cut
sub name        { shift->_value('name',       @_); }

=item  $device->full_name()

=cut
sub full_name   { shift->_value('full_name',  @_); }

=item  $device->description()

=cut
sub description { shift->_value('description',@_); }

=item  $device->control_list()

get a list of the controls

A Control is a group of output/input Channel or any other single element,
that controls the audio in/output
Actualy only input ot output Channels are visible. Switces and other Elements are ignored

B<Return:>

a list of Audio::Alsa::Mixer::Device::Control Objects

B<Example:>

 my @controls=$device->control_list();

=cut
sub control_list
{
  my $self=shift;
  my @list;

  my $cmd=sprintf($controls_cmd,$self->_c_pre('id'));
  my $data=$self->_run_cmd($cmd,'');

  for my $line (split(/\n+/,$data))
  { push(@list,Audio::Alsa::Mixer::Device::Control->_new($self->id(),$1)) if($line=~m#'(.+?)',\d+#); }

  return @list;
}

=item  $device->control(<name>)

get a one control by name

B<Return:>

returns an Audio::Alsa::Mixer::Device::Control Object or undef

B<Example:>

 my $control=$device->control('Master');

=cut
sub control
{
  my $self=shift;
  my $name=shift;
  return undef unless($name);
  for ($self->control_list())
  { return $_ if($_->name() eq $name); }
  return undef;
}

########################################################################
########################################################################
=back

=head2 Audio::Alsa::Mixer::Device::Control

=over 4

=cut
package Audio::Alsa::Mixer::Device::Control;
use base 'Audio::Alsa::Mixer::Helper::Audio';

# new object
sub _new
{
  my $class=shift;
  my $self=bless({},$class);

  $self->{device_id}=shift();
  $self->{name}=shift();
  $self->{channels}=[];

  $self->_read_info();

  return $self;
}

# read infos
sub _read_info
{
  my $self=shift;

  my $cmd=sprintf($channels_cmd, $self->_c_pre('device_id'));
  my $data=$self->_run_cmd($cmd);

  my @channels;

  for my $block (split(/Simple mixer control '/,$data))
  {
    #print "TESTING($self->{name}):\n$block\n##################################\n";
    if($block=~/^$self->{name}/)
    {
      #print "$block\n\n";
      my $pmin=0;
      my $pmax=0;
      if($block=~/Limits:(?:\s+|[^\n]+Playback )(\d+) - (\d+)/)
      { ($pmin,$pmax)=($1,$2); }
      $self->{playback}={min=>$pmin,max=>$pmax};

      my $cmin=0;
      my $cmax=0;
      if($block=~/Limits:(?:\s+|[^\n]+Capture )(\d+) - (\d+)/)
      { ($cmin,$cmax)=($1,$2); }
      $self->{capture}={min=>$cmin,max=>$cmax};


      if($block=~m#Playback channels:\s+([^\n]+)\s*\n#)
      {
        my $cnt=0;
        for(split(/\s+-\s+/,$1))
        {
          $channels[$cnt]->{pos}=$cnt;
          $channels[$cnt]->{name}=$_;
          $channels[$cnt]->{Playback}={ min=>$pmin, max=>$pmax };
          $cnt++;
        }
      }

      if($block=~m#Capture channels:\s+([^\n]+)\s*\n#)
      {
        my $cnt=0;
        for(split(/\s+-\s+/,$1))
        {
          $channels[$cnt]->{pos}=$cnt;
          $channels[$cnt]->{name}=$_;
          $channels[$cnt]->{Capture}={ min=>$cmin, max=>$cmax };
          $cnt++;
        }
      }
      last;
    }
  }

  if(@channels)
  {
    @{$self->{channels}}=();
    for(@channels)
    { push(@{$self->{channels}},Audio::Alsa::Mixer::Device::Control::Channel->_new($self->{device_id},$self->{name},$_)); }
  }
}

=item  $control->name()

=cut
sub name           { shift->_value('name',       @_); }

=item  $control->volume_in_min()

the volume of one input Channel can't be smaler than this value

=cut
sub volume_in_min  { shift->_value('Capture')->{min}; }

=item  $control->volume_in_max()

the volume of one input Channel can't be bigger than this value

=cut
sub volume_in_max  { shift->_value('Capture')->{max}; }

=item  $control->volume_in_max()

the volume of one output Channel can't be smaler than this value

=cut
sub volume_out_min { shift->_value('Playback')->{min}; }

=item  $control->volume_in_max()

the volume of one output Channel can't be bigger than this value

=cut
sub volume_out_max { shift->_value('Playback')->{max}; }

=item  $control->channel_list()

get a list of the channels of this Control

A Channel is a singe input or output Emement

B<Return:>

a list of Audio::Alsa::Mixer::Device::Control::Channel Objects

B<Example:>

 my @channels=$control->channel_list();

=cut
sub channel_list   { return @{shift()->{channels}}; }


=item  $control->channel(<name>)

get one channel of this Control by name

B<Return:>

returns an Audio::Alsa::Mixer::Device::Control::Channel Object or undef

B<Example:>

 my $channel=$control->channel('Left Front');

=cut
sub channel
{
  my $self=shift;
  my $name=shift;
  return undef unless($name);
  for($self->channel_list())
  { return $_ if($_->name() eq $name); }
  return undef;
}

# get/set volume for input/output
sub _volume
{
  my $self=shift;
  my $type=shift;
  return 0 unless($self->{$type} && ref($self->{$type}) eq 'HASH');
  my $val=shift;

  if(!defined($val) || $val eq '%')
  { return $self->_get_volume_list($type,$val); }

  my $perc='';
  if($val=~/^(\d+)%$/)
  {
    $val=int(($self->{$type}->{max}/100)*$1);
    $perc='%';
  }
  elsif($val=~/\D/)
  { return -1; }

  $self->_set_volume_list($type,$val);

  return $self->_get_volume_list($perc)
}

=item  $control->volume_out(<value>)

get/set the ouput Volumes a Control

The value can be an integer value or an percent value as string.

B<Return:>

returns a list of the volumes of every otput Channel in the Control.
If an percent value or an singe "%" as value is set the returnd values are aso in percent

B<Example:>

 # set volume in percent
 my @volumes_percent=$control->volume_out('100%');
 # get volume in percent
 my @volumes_percent=$control->volume_out('%');
 #set volume
 my @volumes=$control->volume_out(0);
 my @volumes=$control->volume_in($control->volume_out_max);
 my @volumes=$control->volume_in($control->volume_out_min);
 #get volume
 my @volumes=$control->volume_out();

=cut
sub volume_out
{
  my $self=shift;
  my $val=shift;
  return $self->_volume('Playback',$val);
}

=item  $control->volume_in(<value>)

get/set the input Volumes a Control

works equal to I<volume_out>

B<Return:>

returns a list of the volumes of every input Channel in the Control

B<Example:>

 my @volumes=$control->volume_in('100%');
 my @volumes=$control->volume_in($control->volume_in_max);

=cut
sub volume_in
{
  my $self=shift;
  my $val=shift;
  return $self->_volume('Capture',$val);
}

=item  $control->mute(<value>)

get/set mute Control

If the value is the string 'toggle' the actual state will be switched.
If the value is 'mute' the Control will muted.
If the value is 'unmute' the Control will unmuted.
Else everey true value will mute the Control unmute otherwise.

B<Return:>

return true/fasle (1/0) if the Control is muted

B<Example:>

 my @volumes=$control->volume_in('100%');
 my @volumes=$control->volume_in($control->volume_in_max);

=cut
sub mute
{
  my $self=shift;
  my $val=shift;

  unless(defined($val))
  {
    my $mute=1;
    my @l=$self->channel_list();
    $mute=$l[0]->mute() if(@l);
    return $mute;
  }

  if($val ne 'toggle' && $val ne 'mute' && $val ne 'unmute')
  {
    if($val)
    { $val='mute'; }
    else
    { $val='unmute'; }
  }

  my $cmd=sprintf($set_mute_cmd,$self->_c_pre('device_id'),$self->{name},$val);
  $self->_exec_cmd($cmd);

  return $self->mute();
}

########################################################################
########################################################################
=back

=head2 Audio::Alsa::Mixer::Device::Control::Channel

=over 4

=cut
package Audio::Alsa::Mixer::Device::Control::Channel;
use base 'Audio::Alsa::Mixer::Helper::Audio';

# create new object
# $opts contains nearly all values
# it's more efficent to runn the amixer control only one time for all Channel
sub _new
{
  my $class=shift;
  my $did=shift();
  my $cname=shift();
  my $opts=shift();
  return bless({ %$opts, device_id=>$did, control_name=>$cname },$class);
}

# get/set volume
sub _volume
{
  my $self=shift;
  my $type=shift;
  return 0 unless($self->{$type} && ref($self->{$type}) eq 'HASH');
  my $val=shift;

  my @l=$self->_get_volume_list($type);
  return -1 if(@l==0 || $#l < $self->{pos} );

  if(!defined($val) || $val eq '%')
  {
    my $vol=$l[$self->{pos}];
    if($vol>-1 && $val && $val eq '%')
    { $vol=int((100/$self->{$type}->{max})*$vol); }
    return $vol;
  }

  my $perc=undef;
  if($val=~/^(\d+)%$/)
  {
    $val=int(($self->{$type}->{max}/100)*$1);
    $perc='%';
  }
  elsif($val=~/\D/)
  { return -1; }

  $l[$self->{pos}]=$val;
  $self->_set_volume_list($type, @l);

  return $self->_volume($type,$perc);
}

=item  $channel->mute()

return tue if Channes is muted

B<Return:>

return true/fasle (1/0) if the Channel is muted

B<Example:>

 print 'Channel ',$channel->name(),' is ',$channel->mute()?'':'not ',"muted \n";

=cut
sub mute
{
  my $self=shift;
  my $mute=0;

  my $cmd=sprintf($get_mute_cmd, $self->_c_pre('device_id'), $self->{control_name});
  my $data=$self->_run_cmd($cmd,'');

  my $n=$self->{name};
  if($data=~/\Q$n\E:[^\n]+\[(on|off)\]/)
  { $mute=$1 eq 'on'?0:1; }

  return $mute;
}

=item  $channel->mute()

Channel Name

=cut
sub name        { shift->_value('name',       @_); }

=item  $channel->volume_out()

get/set volume for output

works like volume_out in Alsa::Mixer::Controler

=cut
sub volume_out  { shift->_volume('Playback',  @_); }

=item  $channel->volume_in()

get/set volume for input

works like volume_in in Alsa::Mixer::Controler

=cut
sub volume_in   { shift->_volume('Capture',  @_); }

=item  $channel->volume_in_min()

=cut
sub volume_in_min   { shift->_value('Capture')->{min}; }

=item  $channel->volume_in_max()

=cut
sub volume_in_max   { shift->_value('Capture')->{max}; }

=item  $channel->volume_out_min()

=cut
sub volume_out_min  { shift->_value('Playback')->{min}; }

=item  $channel->volume_out_max()

=cut
sub volume_out_max  { shift->_value('Playback')->{max}; }

=back

=head1 AUTHOR

Tobias Grönhagen <osm@topeg.de>

=head1 VERSION

Version 0.1 2 Feb 2011

=head1 COPYRIGHT

Copyright (c) Tobias Grönhagen. All rights reserved.

=head1 LICENSE

This program is free software; you can redistribute it
and/or modify it under the same terms and conditions as
Perl itself.

This means that you can, at your option, redistribute it and/or modify it under
either the terms the GNU Public License (GPL) version 2 or later, or under the
Perl Artistic License.

See http://dev.perl.org/licenses/

=head1 DISCLAIMER

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.

Use of this software in any way or in any form, source or binary,
is not allowed in any country which prohibits disclaimers of any
implied warranties of merchantability or fitness for a particular
purpose or any disclaimers of a similar nature.

IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION (INCLUDING, BUT NOT
LIMITED TO, LOST PROFITS) EVEN IF I HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE

=cut
1;
